from pathlib import Path

from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from .classes import DefinedStorageLazy
from .events import event_download_file_created, event_download_file_deleted
from .literals import (
    STORAGE_NAME_DOWNLOAD_FILE, STORAGE_NAME_SHARED_UPLOADED_FILE
)
from .managers import DownloadFileManager, SharedUploadedFileManager
from .model_mixins import (
    DatabaseFileModelMixin, DownloadFileBusinessLogicMixin
)
from .utils import download_file_upload_to, shared_uploaded_file_upload_to


class DownloadFile(
    DatabaseFileModelMixin, DownloadFileBusinessLogicMixin,
    ExtraDataModelMixin, models.Model
):
    """
    Keep a database link to a stored file. Used for generated files meant
    to be downloaded at a later time.
    """

    file = models.FileField(
        storage=DefinedStorageLazy(
            name=STORAGE_NAME_DOWNLOAD_FILE
        ), upload_to=download_file_upload_to, verbose_name=_('File')
    )
    label = models.CharField(
        db_index=True, max_length=192, verbose_name=_('Label')
    )
    user = models.ForeignKey(
        editable=False, on_delete=models.CASCADE,
        related_name='download_files', to=settings.AUTH_USER_MODEL,
        verbose_name=_('User')
    )

    objects = DownloadFileManager()

    class Meta:
        ordering = ('-datetime',)
        verbose_name = _('Download file')
        verbose_name_plural = _('Download files')

    def __str__(self):
        # Ensure the returned value is not a promise.
        return str(self.filename or self.label)

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_download_file_deleted
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse(viewname='storage:download_file_list')

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_download_file_created,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)


class SharedUploadedFile(DatabaseFileModelMixin, models.Model):
    """
    Keep a database link to a stored file. Used to share files between code
    that runs out of process.
    """

    file = models.FileField(
        storage=DefinedStorageLazy(
            name=STORAGE_NAME_SHARED_UPLOADED_FILE
        ), upload_to=shared_uploaded_file_upload_to, verbose_name=_('File')
    )
    filename = models.CharField(
        blank=True, max_length=255, verbose_name=_('Filename')
    )

    objects = SharedUploadedFileManager()

    class Meta:
        verbose_name = _('Shared uploaded file')
        verbose_name_plural = _('Shared uploaded files')

    def __str__(self):
        return self.filename

    def save(self, *args, **kwargs):
        self.filename = self.filename or Path(path=self.file.name).name
        super().save(*args, **kwargs)
        self.file.close()



from django.core.exceptions import PermissionDenied

from rest_framework.permissions import BasePermission

from mayan.apps.acls.models import AccessControlList
from mayan.apps.permissions import Permission


class MayanPermission(BasePermission):
    def get_mayan_object_permissions(self, request, view):
        try:
            method_get_mayan_object_permissions = getattr(
                view, 'get_mayan_object_permissions'
            )
        except AttributeError:
            return getattr(
                view, 'mayan_object_permissions', {}
            ).get(request.method, None)
        else:
            return method_get_mayan_object_permissions(
                request=request, view=view
            )

    def get_mayan_view_permissions(self, request, view):
        try:
            method_get_mayan_view_permissions = getattr(
                view, 'get_mayan_view_permissions'
            )
        except AttributeError:
            return getattr(
                view, 'mayan_view_permissions', {}
            ).get(request.method, None)
        else:
            return method_get_mayan_view_permissions(
                request=request, view=view
            )

    def has_object_permission(self, request, view, obj):
        permissions = self.get_mayan_object_permissions(
            request=request, view=view
        )
        user = request.user

        if permissions:
            if user.is_authenticated:
                try:
                    AccessControlList.objects.check_access(
                        obj=obj, permissions=permissions,
                        user=user
                    )
                except PermissionDenied:
                    return False
                else:
                    return True
            else:
                return False
        else:
            return True

    def has_permission(self, request, view):
        permissions = self.get_mayan_view_permissions(
            request=request, view=view
        )
        user = request.user

        if permissions:
            if user.is_authenticated:
                try:
                    Permission.check_user_permissions(
                        permissions=permissions, user=user
                    )
                except PermissionDenied:
                    return False
                else:
                    return True
            else:
                return False
        else:
            return True



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0028_workflowstateescalation')
    ]

    operations = [
        migrations.RenameField(
            model_name='workflowstateaction',
            old_name='action_data',
            new_name='backend_data'
        ),
        migrations.RenameField(
            model_name='workflowstateaction',
            old_name='action_path',
            new_name='backend_path'
        )
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_b, worker_c

queue_indexing = CeleryQueue(
    label=_('Indexing'), name='indexing', worker=worker_b
)
queue_indexing_slow = CeleryQueue(
    label=_('Indexing slow'), name='indexing_slow', worker=worker_c
)

queue_indexing.add_task_type(
    label=_('Remove document'),
    dotted_path='mayan.apps.document_indexing.tasks.task_index_instance_document_remove'
)
queue_indexing.add_task_type(
    label=_('Index document'),
    dotted_path='mayan.apps.document_indexing.tasks.task_index_instance_document_add'
)

queue_indexing_slow.add_task_type(
    label=_('Rebuild index'),
    dotted_path='mayan.apps.document_indexing.tasks.task_index_template_rebuild'
)



from django.conf.urls import url

from .api_views import (
    APIStoredCredentialListView, APIStoredCredentialDetailView
)
from .views import (
    StoredCredentialBackendSelectionView, StoredCredentialCreateView,
    StoredCredentialDeleteView, StoredCredentialEditView,
    StoredCredentialListView
)

urlpatterns = [
    url(
        regex=r'^credentials/backend/selection/$',
        name='stored_credential_backend_selection',
        view=StoredCredentialBackendSelectionView.as_view()
    ),
    url(
        regex=r'^credentials/(?P<backend_path>[a-zA-Z0-9_.]+)/create/$',
        name='stored_credential_create',
        view=StoredCredentialCreateView.as_view()
    ),
    url(
        regex=r'^credentials/(?P<stored_credential_id>\d+)/delete/$',
        name='stored_credential_delete',
        view=StoredCredentialDeleteView.as_view()
    ),
    url(
        regex=r'^credentials/(?P<stored_credential_id>\d+)/edit/$',
        name='stored_credential_edit', view=StoredCredentialEditView.as_view()
    ),
    url(
        regex=r'^credentials/$', name='stored_credential_list',
        view=StoredCredentialListView.as_view()
    )
]

api_urls = [
    url(
        regex=r'^credentials/$', name='credential-list',
        view=APIStoredCredentialListView.as_view()
    ),
    url(
        regex=r'^credentials/(?P<credential_id>[0-9]+)/$',
        name='credential-detail', view=APIStoredCredentialDetailView.as_view()
    )
]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_LOCK_MANAGER_BACKEND, DEFAULT_LOCK_MANAGER_BACKEND_ARGUMENTS,
    DEFAULT_LOCK_MANAGER_DEFAULT_LOCK_TIMEOUT
)

namespace = SettingNamespace(
    label=_('Lock manager'), name='lock_manager'
)

setting_backend = namespace.add_setting(
    default=DEFAULT_LOCK_MANAGER_BACKEND, global_name='LOCK_MANAGER_BACKEND',
    help_text=_(
        'Path to the class to use when to request and release '
        'resource locks.'
    )
)
setting_backend_arguments = namespace.add_setting(
    default=DEFAULT_LOCK_MANAGER_BACKEND_ARGUMENTS,
    global_name='LOCK_MANAGER_BACKEND_ARGUMENTS', help_text=_(
        'Arguments to pass to the LOCK_MANAGER_BACKEND.'
    )
)
setting_default_lock_timeout = namespace.add_setting(
    default=DEFAULT_LOCK_MANAGER_DEFAULT_LOCK_TIMEOUT,
    global_name='LOCK_MANAGER_DEFAULT_LOCK_TIMEOUT', help_text=_(
        'Default amount of time in seconds after which a resource '
        'lock will be automatically released.'
    )
)



from mayan.apps.documents.permissions import permission_document_create

from ..permissions import permission_sources_edit, permission_sources_view
from ..source_backend_actions.base import SourceBackendAction
from ..source_backend_actions.interfaces import (
    SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestView, SourceBackendActionInterfaceTask
)
from ..source_backend_actions.mixins.callback_mixins import SourceBackendActionMixinCallbackDocumentUploadBase
from ..source_backend_actions.mixins.document_mixins import SourceBackendActionMixinDocumentUploadInteractive
from ..source_backend_actions.mixins.document_type_mixins import SourceBackendActionMixinDocumentTypeInteractive

from .literals import (
    TEST_SOURCE_ACTION_CONFIRM_FALSE_NAME,
    TEST_SOURCE_ACTION_CONFIRM_TRUE_NAME
)
from .source_backend_mixins import SourceBackendActionMixinFileUser


class SourceBackendActionDocumentUploadBasicInteractive(
    SourceBackendActionMixinDocumentUploadInteractive,
    SourceBackendActionMixinCallbackDocumentUploadBase,
    SourceBackendActionMixinFileUser,
    SourceBackendActionMixinDocumentTypeInteractive, SourceBackendAction
):
    name = 'document_upload'
    permission = permission_document_create


class SourceBackendActionTestConfirmFalse(SourceBackendAction):
    confirmation = False
    name = TEST_SOURCE_ACTION_CONFIRM_FALSE_NAME
    permission = permission_sources_view

    class Interface:
        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            """Empty interface. No arguments or processing."""

        class Task(SourceBackendActionInterfaceTask):
            """Empty interface. No arguments or processing."""

        class View(SourceBackendActionInterfaceRequestView):
            """Empty interface. No arguments or processing."""


class SourceBackendActionTestConfirmTrue(SourceBackendAction):
    confirmation = True
    name = TEST_SOURCE_ACTION_CONFIRM_TRUE_NAME
    permission = permission_sources_edit

    class Interface:
        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            """Empty interface. No arguments or processing."""

        class Task(SourceBackendActionInterfaceTask):
            """Empty interface. No arguments or processing."""

        class View(SourceBackendActionInterfaceRequestView):
            """Empty interface. No arguments or processing."""



from django.contrib import admin

from .models import DocumentTypeOCRSettings, DocumentVersionPageOCRContent


@admin.register(DocumentTypeOCRSettings)
class DocumentTypeOCRSettingsAdmin(admin.ModelAdmin):
    list_display = ('document_type', 'auto_ocr')


@admin.register(DocumentVersionPageOCRContent)
class DocumentVersionPageOCRContentAdmin(admin.ModelAdmin):
    list_display = ('document_version_page',)



import logging

from django.core.files.storage import Storage
from django.db import migrations
from django.utils.module_loading import import_string

STORAGE_PATH_UPDATE_MAP = {
    'mayan.apps.document_states.storages.storage_workflowimagecache': 'mayan.apps.document_states.storages.storage_workflow_image',
    'mayan.apps.documents.storages.storage_documentimagecache': 'mayan.apps.documents.storages.storage_document_image_cache'
}
logger = logging.getLogger(name=__name__)


class DummyStorage(Storage):
    def delete(self, name):
        """
        Do nothing. This dummy storage avoids an if in the
        `cache_partition_file` loop.
        """


def code_purge_and_delete_caches(apps, schema_editor):
    Cache = apps.get_model(
        app_label='file_caching', model_name='Cache'
    )
    CachePartitionFile = apps.get_model(
        app_label='file_caching', model_name='CachePartitionFile'
    )

    cursor_primary = schema_editor.connection.create_cursor(
        name='file_caching_partition_files'
    )
    cursor_secondary = schema_editor.connection.cursor()

    query = str(
        CachePartitionFile.objects.values(
            'partition__name', 'filename', 'partition__cache_id'
        ).query
    )

    '''
    Expected resulting query:
        SELECT
            "file_caching_cachepartition"."name",
            "file_caching_cachepartitionfile"."filename",
            "file_caching_cachepartition"."cache_id"
        FROM "file_caching_cachepartitionfile"
        INNER JOIN
            "file_caching_cachepartition" ON (
                "file_caching_cachepartitionfile"."partition_id" = "file_caching_cachepartition"."id"
            )
    '''

    cache_storages = {}
    for cache in Cache.objects.using(alias=schema_editor.connection.alias).all():
        try:
            cache_storages[cache.pk] = import_string(
                dotted_path=cache.storage_instance_path
            ).get_storage_instance()
        except ImportError as exception:
            logger.error(
                'Storage "%s" not found. Remove the files from this '
                'storage manually.; %s', cache.storage_instance_path,
                exception
            )
            cache_storages[cache.pk] = DummyStorage()

    cursor_primary.execute(query)
    for partition_name, filename, cache_id in cursor_primary.fetchall():
        cache_storages[
            cache_id
        ].delete(
            name='{}-{}'.format(
                partition_name,
                filename
            )
        )

    table_names = (
        'file_caching_cachepartitionfile', 'file_caching_cachepartition',
        'file_caching_cache'
    )

    for table_name in table_names:
        cursor_secondary.execute(
            'DELETE FROM {};'.format(
                schema_editor.connection.ops.quote_name(
                    name=table_name
                )
            )
        )


def code_update_storage_paths(apps, schema_editor):
    Cache = apps.get_model(
        app_label='file_caching', model_name='Cache'
    )

    for cache in Cache.objects.using(alias=schema_editor.connection.alias).all():
        cache.storage_instance_path = STORAGE_PATH_UPDATE_MAP.get(
            cache.storage_instance_path, cache.storage_instance_path
        )
        cache.save()


def code_update_storage_paths_reverse(apps, schema_editor):
    storage_path_update_map_inverted = {
        value: key for key, value in STORAGE_PATH_UPDATE_MAP.items()
    }

    Cache = apps.get_model(
        app_label='file_caching', model_name='Cache'
    )

    for cache in Cache.objects.using(alias=schema_editor.connection.alias).all():
        cache.storage_instance_path = storage_path_update_map_inverted.get(
            cache.storage_instance_path, cache.storage_instance_path
        )
        cache.save()


class Migration(migrations.Migration):
    dependencies = [
        ('file_caching', '0004_auto_20200309_0922')
    ]
    operations = [
        migrations.RunPython(
            code=code_update_storage_paths,
            reverse_code=code_update_storage_paths_reverse
        ),
        migrations.RunPython(
            code=code_purge_and_delete_caches,
            reverse_code=migrations.RunPython.noop
        )
    ]



COMMENT_APP_TEMPLATE_CACHE_DISABLE = '{# appearance_app_template_nocache #}'

COMMAND_NAME_APPEARANCE_PREPARE_STATIC = 'appearance_prepare_static'

DEFAULT_AJAX_REDIRECTION_CODE = 278
DEFAULT_MAXIMUM_TITLE_LENGTH = 120
DEFAULT_MESSAGE_POSITION = 'top-right'
DEFAULT_MENU_POLLING_INTERVAL = 5000
DEFAULT_THROTTLING_MAXIMUM_REQUESTS = 10
DEFAULT_THROTTLING_TIMEOUT = 5000



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.db import models

from mayan.apps.events.classes import EventModelRegistry
from mayan.apps.testing.tests.base import BaseTestCase

from ..classes import ModelPermission
from ..models import AccessControlList

from .mixins import ACLTestMixin


class PermissionTestCase(ACLTestMixin, BaseTestCase):
    auto_create_acl_test_object = False

    def test_check_access_without_permissions(self):
        self._create_acl_test_object()

        with self.assertRaises(expected_exception=PermissionDenied):
            AccessControlList.objects.check_access(
                obj=self._test_object,
                permissions=(self._test_permission,),
                user=self._test_case_user,
            )

    def test_filtering_without_permissions(self):
        self._create_acl_test_object()

        self.assertEqual(
            AccessControlList.objects.restrict_queryset(
                permission=self._test_permission,
                queryset=self._test_object._meta.model._default_manager.all(),
                user=self._test_case_user
            ).count(), 0
        )

    def test_check_access_with_acl(self):
        self._create_acl_test_object()

        self.grant_access(
            obj=self._test_object, permission=self._test_permission
        )

        try:
            AccessControlList.objects.check_access(
                obj=self._test_object, permissions=(self._test_permission,),
                user=self._test_case_user,
            )
        except PermissionDenied:
            self.fail('PermissionDenied exception was not expected.')

    def test_filtering_with_permissions(self):
        self._create_acl_test_object()

        self.grant_access(
            obj=self._test_object, permission=self._test_permission
        )

        self.assertTrue(
            self._test_object in AccessControlList.objects.restrict_queryset(
                permission=self._test_permission,
                queryset=self._test_object._meta.model._default_manager.all(),
                user=self._test_case_user
            )
        )

    def _setup_child_parent_test_objects(self):
        self._create_test_permission()
        self.TestModelParent = self._create_test_model(
            model_name='TestModelParent'
        )
        self.TestModelChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent',
                )
            }, model_name='TestModelChild'
        )

        ModelPermission.register(
            model=self.TestModelParent, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register(
            model=self.TestModelChild, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register_inheritance(
            model=self.TestModelChild, related='parent',
        )

        self._test_object_parent = self.TestModelParent.objects.create()
        self._test_object_child = self.TestModelChild.objects.create(
            parent=self._test_object_parent
        )

    def test_check_access_with_inherited_acl(self):
        self._setup_child_parent_test_objects()

        self.grant_access(
            obj=self._test_object_parent, permission=self._test_permission
        )

        try:
            AccessControlList.objects.check_access(
                obj=self._test_object_child,
                permissions=(self._test_permission,),
                user=self._test_case_user
            )
        except PermissionDenied:
            self.fail('PermissionDenied exception was not expected.')

    def test_check_access_with_inherited_acl_and_local_acl(self):
        self._setup_child_parent_test_objects()

        self.grant_access(
            obj=self._test_object_parent, permission=self._test_permission
        )
        self.grant_access(
            obj=self._test_object_child, permission=self._test_permission
        )

        try:
            AccessControlList.objects.check_access(
                obj=self._test_object_child,
                permissions=(self._test_permission,),
                user=self._test_case_user
            )
        except PermissionDenied:
            self.fail('PermissionDenied exception was not expected.')

    def test_filtering_with_inherited_permissions(self):
        self._setup_child_parent_test_objects()

        self.grant_access(
            obj=self._test_object_parent, permission=self._test_permission
        )

        result = AccessControlList.objects.restrict_queryset(
            permission=self._test_permission,
            queryset=self._test_object_child._meta.model._default_manager.all(),
            user=self._test_case_user
        )
        self.assertTrue(self._test_object_child in result)

    def test_filtering_with_inherited_permissions_and_local_acl(self):
        self._setup_child_parent_test_objects()

        self.grant_permission(permission=self._test_permission)
        self.grant_access(
            obj=self._test_object_parent, permission=self._test_permission
        )
        self.grant_access(
            obj=self._test_object_child, permission=self._test_permission
        )

        result = AccessControlList.objects.restrict_queryset(
            permission=self._test_permission,
            queryset=self._test_object_child._meta.model._default_manager.all(),
            user=self._test_case_user,
        )
        self.assertTrue(self._test_object_child in result)

    def test_method_get_absolute_url(self):
        self._create_acl_test_object()
        self._create_test_acl()

        self.assertTrue(
            self._test_acl.get_absolute_url()
        )


class InheritedPermissionTestCase(ACLTestMixin, BaseTestCase):
    def test_retrieve_inherited_role_permission_not_model_applicable(self):
        self.TestModel = self._create_test_model()
        EventModelRegistry.register(model=self.TestModel)
        self._test_object = self.TestModel.objects.create()
        self._create_test_acl()
        self._create_test_permission()

        self._test_role.grant(permission=self._test_permission)

        queryset = AccessControlList.objects.get_inherited_permissions(
            obj=self._test_object, role=self._test_role
        )
        self.assertTrue(
            self._test_permission.stored_permission not in queryset
        )

    def test_retrieve_inherited_role_permission_model_applicable(self):
        self.TestModel = self._create_test_model()
        EventModelRegistry.register(model=self.TestModel)
        self._test_object = self.TestModel.objects.create()
        self._create_test_acl()
        self._create_test_permission()

        ModelPermission.register(
            model=self._test_object._meta.model, permissions=(
                self._test_permission,
            )
        )
        self._test_role.grant(permission=self._test_permission)

        queryset = AccessControlList.objects.get_inherited_permissions(
            obj=self._test_object, role=self._test_role
        )
        self.assertTrue(self._test_permission.stored_permission in queryset)

    def test_retrieve_inherited_related_parent_child_permission(self):
        self._create_test_permission()

        self.TestModelParent = self._create_test_model(
            model_name='TestModelParent'
        )
        self.TestModelChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent',
                )
            }, model_name='TestModelChild'
        )

        ModelPermission.register(
            model=self.TestModelParent, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register(
            model=self.TestModelChild, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register_inheritance(
            model=self.TestModelChild, related='parent',
        )

        parent = self.TestModelParent.objects.create()
        child = self.TestModelChild.objects.create(parent=parent)

        AccessControlList.objects.grant(
            obj=parent, permission=self._test_permission,
            role=self._test_role
        )
        queryset = AccessControlList.objects.get_inherited_permissions(
            obj=child, role=self._test_role
        )

        self.assertTrue(self._test_permission.stored_permission in queryset)

    def test_retrieve_inherited_related_grandparent_parent_child_permission(
        self
    ):
        self._create_test_permission()

        self.TestModelGrandParent = self._create_test_model(
            model_name='TestModelGrandParent'
        )
        self.TestModelParent = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelGrandParent',
                )
            }, model_name='TestModelParent'
        )
        self.TestModelChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent',
                )
            }, model_name='TestModelChild'
        )

        ModelPermission.register(
            model=self.TestModelGrandParent, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register(
            model=self.TestModelParent, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register(
            model=self.TestModelChild, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register_inheritance(
            model=self.TestModelChild, related='parent',
        )
        ModelPermission.register_inheritance(
            model=self.TestModelParent, related='parent',
        )

        grandparent = self.TestModelGrandParent.objects.create()
        parent = self.TestModelParent.objects.create(parent=grandparent)
        child = self.TestModelChild.objects.create(parent=parent)

        AccessControlList.objects.grant(
            obj=grandparent, permission=self._test_permission,
            role=self._test_role
        )

        queryset = AccessControlList.objects.get_inherited_permissions(
            obj=child, role=self._test_role
        )

        self.assertTrue(self._test_permission.stored_permission in queryset)

    def test_sub_model_with_multiple_inheritance_parent_with_common_super_parent_paths(self):
        self._create_test_permission()

        self.TestModelLevel1 = self._create_test_model(
            model_name='TestModelLevel1'
        )
        self.TestModelLevel2 = self._create_test_model(
            fields={
                'level_1': models.ForeignKey(
                    on_delete=models.CASCADE, to='TestModelLevel1'
                )
            }, model_name='TestModelLevel2'
        )
        self.TestModelLevel3 = self._create_test_model(
            fields={
                'level_2': models.ForeignKey(
                    on_delete=models.CASCADE, to='TestModelLevel2'
                )
            }, model_name='TestModelLevel3'
        )
        self.TestModelLevel4 = self._create_test_model(
            fields={
                'level_3': models.ForeignKey(
                    on_delete=models.CASCADE, to='TestModelLevel3'
                )
            }, model_name='TestModelLevel4'
        )

        ModelPermission.register(
            model=self.TestModelLevel2, permissions=(
                self._test_permission,
            )
        )

        ModelPermission.register_inheritance(
            model=self.TestModelLevel2, related='level_1',
        )
        ModelPermission.register_inheritance(
            model=self.TestModelLevel3, related='level_2__level_1',
        )
        ModelPermission.register_inheritance(
            model=self.TestModelLevel3, related='level_2',
        )
        ModelPermission.register_inheritance(
            model=self.TestModelLevel4, related='level_3',
        )

        level_1 = self.TestModelLevel1.objects.create()
        level_2 = self.TestModelLevel2.objects.create(level_1=level_1)
        level_3 = self.TestModelLevel3.objects.create(level_2=level_2)
        level_4 = self.TestModelLevel4.objects.create(level_3=level_3)

        self.grant_access(
            obj=level_2, permission=self._test_permission
        )

        queryset = AccessControlList.objects.restrict_queryset(
            permission=self._test_permission,
            queryset=self.TestModelLevel4.objects.all(),
            user=self._test_case_user
        )

        self.assertTrue(level_4 in queryset)


class GenericForeignKeyFieldModelTestCase(ACLTestMixin, BaseTestCase):
    auto_create_acl_test_object = False

    def test_generic_foreign_key_model_with_alternate_ct_and_fk(self):
        self._create_test_permission()

        self.TestModelExternal = self._create_test_model(
            model_name='TestModelExternal'
        )
        self.TestModelChild = self._create_test_model(
            fields={
                'content_type_1': models.ForeignKey(
                    on_delete=models.CASCADE,
                    related_name='object_content_type',
                    to=ContentType
                ),
                'object_id_1': models.PositiveIntegerField(),
                'content_object_1': GenericForeignKey(
                    ct_field='content_type_1', fk_field='object_id_1',
                )
            }, model_name='TestModelChild'
        )

        ModelPermission.register(
            model=self.TestModelExternal, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register(
            model=self.TestModelChild, permissions=(
                self._test_permission,
            )
        )

        ModelPermission.register_inheritance(
            model=self.TestModelChild, related='content_object_1'
        )

        test_external_object = self.TestModelExternal.objects.create()
        test_object = self.TestModelChild.objects.create(
            content_object_1=test_external_object
        )

        self.grant_access(
            obj=test_external_object, permission=self._test_permission
        )

        queryset = AccessControlList.objects.restrict_queryset(
            queryset=self.TestModelChild.objects.all(),
            permission=self._test_permission, user=self._test_case_user
        )

        self.assertTrue(test_object in queryset)

    def test_generic_foreign_key_model_with_multiple_alternate_ct_and_fk(self):
        self._create_test_permission()

        self.TestModelExternal = self._create_test_model(
            model_name='TestModelExternal'
        )
        self.TestModelChild = self._create_test_model(
            fields={
                'content_type_1': models.ForeignKey(
                    on_delete=models.CASCADE,
                    related_name='object_content_type', to=ContentType
                ),
                'object_id_1': models.PositiveIntegerField(),
                'content_object_1': GenericForeignKey(
                    ct_field='content_type_1', fk_field='object_id_1',
                ),
                'content_type_2': models.ForeignKey(
                    blank=True, null=True, on_delete=models.CASCADE,
                    related_name='object_content_type',
                    to=ContentType
                ),
                'object_id_2': models.PositiveIntegerField(
                    blank=True, null=True
                ),
                'content_object_2': GenericForeignKey(
                    ct_field='content_type_2', fk_field='object_id_2',
                )
            }, model_name='TestModelChild'
        )

        ModelPermission.register(
            model=self.TestModelExternal, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register(
            model=self.TestModelChild, permissions=(
                self._test_permission,
            )
        )

        ModelPermission.register_inheritance(
            model=self.TestModelChild, related='content_object_1',
        )
        ModelPermission.register_inheritance(
            model=self.TestModelChild, related='content_object_2',
        )

        test_external_object = self.TestModelExternal.objects.create()
        test_object = self.TestModelChild.objects.create(
            content_object_1=test_external_object
        )

        self.grant_access(
            obj=test_external_object, permission=self._test_permission
        )

        queryset = AccessControlList.objects.restrict_queryset(
            queryset=self.TestModelChild.objects.all(),
            permission=self._test_permission, user=self._test_case_user
        )

        self.assertTrue(test_object in queryset)

    def test_generic_foreign_key_model_with_typecasting(self):
        self._create_test_permission()

        self.TestModelExternal = self._create_test_model(
            model_name='TestModelExternal'
        )
        self.TestModelChild = self._create_test_model(
            fields={
                'content_type': models.ForeignKey(
                    on_delete=models.CASCADE,
                    related_name='object_content_type',
                    to=ContentType
                ),
                'object_id': models.CharField(max_length=255),
                'content_object': GenericForeignKey(
                    ct_field='content_type', fk_field='object_id',
                )
            }, model_name='TestModelChild'
        )

        ModelPermission.register(
            model=self.TestModelExternal, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register(
            model=self.TestModelChild, permissions=(
                self._test_permission,
            )
        )

        ModelPermission.register_inheritance(
            model=self.TestModelChild, related='content_object',
            fk_field_cast=models.CharField
        )

        test_external_object = self.TestModelExternal.objects.create()
        test_object = self.TestModelChild.objects.create(
            content_object=test_external_object
        )

        self.grant_access(
            obj=test_external_object, permission=self._test_permission
        )

        queryset = AccessControlList.objects.restrict_queryset(
            queryset=self.TestModelChild.objects.all(),
            permission=self._test_permission, user=self._test_case_user
        )

        self.assertTrue(test_object in queryset)


class ProxyModelPermissionTestCase(ACLTestMixin, BaseTestCase):
    def test_proxy_model_filtering_no_permission(self):
        self._create_acl_test_object_base()
        self._create_acl_test_object_proxy()

        proxy_object = self.TestModelProxy.objects.get(pk=self._test_object.pk)

        self.assertFalse(
            proxy_object in AccessControlList.objects.restrict_queryset(
                permission=self._test_permission,
                queryset=self.TestModelProxy.objects.all(),
                user=self._test_case_user
            )
        )

    def test_proxy_model_filtering_with_access(self):
        self._create_acl_test_object_base()
        self._create_acl_test_object_proxy()

        self.grant_access(
            obj=self._test_object, permission=self._test_permission
        )

        proxy_object = self.TestModelProxy.objects.get(pk=self._test_object.pk)

        self.assertTrue(
            proxy_object in AccessControlList.objects.restrict_queryset(
                permission=self._test_permission,
                queryset=self.TestModelProxy.objects.all(),
                user=self._test_case_user
            )
        )

    def test_proxy_model_inheritance_with_access(self):
        self._create_test_permission()

        self.TestModelParent = self._create_test_model(
            model_name='TestModelParent'
        )
        self.TestModelChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent',
                )
            }, model_name='TestModelChild'
        )
        self.TestModelProxy = self._create_test_model(
            base_class=self.TestModelChild, model_name='TestModelProxy',
            options={
                'proxy': True
            }
        )

        ModelPermission.register(
            model=self.TestModelParent, permissions=(
                self._test_permission,
            )
        )
        ModelPermission.register_inheritance(
            model=self.TestModelChild, related='parent',
        )

        parent = self.TestModelParent.objects.create()
        child = self.TestModelChild.objects.create(parent=parent)

        self.grant_access(
            obj=parent, permission=self._test_permission
        )

        proxy_object = self.TestModelProxy.objects.get(pk=child.pk)

        self.assertTrue(
            proxy_object in AccessControlList.objects.restrict_queryset(
                permission=self._test_permission,
                queryset=self.TestModelProxy.objects.all(),
                user=self._test_case_user
            )
        )



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_document_created, event_document_edited,
    event_document_file_created, event_document_file_edited,
    event_document_type_changed, event_document_version_created,
    event_document_version_page_created
)
from ..models.document_models import Document
from ..permissions import (
    permission_document_change_type, permission_document_create,
    permission_document_properties_edit, permission_document_view
)

from .mixins.document_mixins import (
    DocumentAPIViewTestMixin, DocumentTestMixin
)


class DocumentAPIViewTestCase(DocumentAPIViewTestMixin, BaseAPITestCase):
    auto_upload_test_document = False

    def test_document_create_api_view_no_permission(self):
        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_document_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_create_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_document_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

    def test_document_detail_api_view_no_permission(self):
        self._create_test_document_stub()

        self._clear_events()

        response = self._request_test_document_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_detail_api_view_with_access(self):
        self._create_test_document_stub()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['id'], self._test_document.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_detail_api_view_with_access(self):
        self._create_test_document_stub()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_edit_via_patch_api_view_no_permission(self):
        self._create_test_document_stub()

        document_description = self._test_document.description

        self._clear_events()

        response = self._request_test_document_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()
        self.assertEqual(
            self._test_document.description, document_description
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_edit_via_patch_api_view_with_access(self):
        self._create_test_document_stub()

        document_description = self._test_document.description

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_properties_edit
        )

        self._clear_events()

        response = self._request_test_document_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document.refresh_from_db()
        self.assertNotEqual(
            self._test_document.description, document_description
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_edited.id)

    def test_trashed_document_edit_via_patch_api_view_with_access(self):
        self._create_test_document_stub()

        document_description = self._test_document.description

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_properties_edit
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()
        self.assertEqual(
            self._test_document.description, document_description
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_edit_via_put_api_view_no_permission(self):
        self._create_test_document_stub()

        document_description = self._test_document.description

        self._clear_events()

        response = self._request_test_document_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()
        self.assertEqual(
            self._test_document.description, document_description
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_edit_via_put_api_view_with_access(self):
        self._create_test_document_stub()

        document_description = self._test_document.description

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_properties_edit
        )

        self._clear_events()

        response = self._request_test_document_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document.refresh_from_db()
        self.assertNotEqual(
            self._test_document.description, document_description
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_edited.id)

    def test_trashed_document_edit_via_put_api_view_with_access(self):
        self._create_test_document_stub()

        document_description = self._test_document.description

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_properties_edit
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()
        self.assertEqual(
            self._test_document.description, document_description
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_list_api_view_no_permission(self):
        self._create_test_document_stub()

        self._clear_events()

        response = self._request_test_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_list_api_view_with_access(self):
        self._create_test_document_stub()

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['results'][0]['id'], self._test_document.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_list_api_view_with_access(self):
        self._create_test_document_stub()

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_upload_api_view_no_permission(self):
        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_document_upload_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_upload_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_document_upload_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(
            self._test_document.pk, response.data['id']
        )
        self.assertEqual(
            self._test_document.label,
            self._test_document.file_latest.filename
        )
        self.assertEqual(self._test_document.file_latest.size, 17436)
        self.assertEqual(
            self._test_document.file_latest.mimetype, 'image/png'
        )
        self.assertEqual(self._test_document.file_latest.encoding, 'binary')
        self.assertEqual(
            self._test_document.file_latest.checksum,
            'efa10e6cc21f83078aaa94d5cbe51de67b51af706143bafc7fd6d4c02124879a'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        # Document created

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        # Document file created

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document.file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        # Document file edited (MIME type, page count update)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, self._test_document.file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        # Document version created

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(
            events[3].target, self._test_document.version_active
        )
        self.assertEqual(events[3].verb, event_document_version_created.id)

        # Document version page created

        self.assertEqual(
            events[4].actor, self._test_case_user
        )
        self.assertEqual(
            events[4].action_object, self._test_document.version_active
        )
        self.assertEqual(
            events[4].target,
            self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )


class DocumentChangeTypeAPIViewTestCase(
    DocumentAPIViewTestMixin, DocumentTestMixin, BaseAPITestCase
):
    auto_create_test_document_stub = True

    def setUp(self):
        super().setUp()
        self._create_test_document_type()

    def test_document_change_type_api_view_no_permission(self):
        test_document_type = self._test_document.document_type

        self._clear_events()

        response = self._request_test_document_change_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()
        self.assertEqual(
            self._test_document.document_type, test_document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_change_type_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_change_type
        )

        test_document_type = self._test_document.document_type

        self._clear_events()

        response = self._request_test_document_change_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        error_dict = response.json()
        self.assertTrue('document_type_id' in error_dict)
        self.assertTrue(
            'object does not exist' in error_dict['document_type_id'][0]
        )

        self._test_document.refresh_from_db()
        self.assertEqual(
            self._test_document.document_type, test_document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_change_type_api_view_with_document_type_access(self):
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_change_type
        )

        test_document_type = self._test_document.document_type

        self._clear_events()

        response = self._request_test_document_change_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()
        self.assertEqual(
            self._test_document.document_type, test_document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_change_type_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_change_type
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_change_type
        )

        test_document_type = self._test_document.document_type

        self._clear_events()

        response = self._request_test_document_change_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document.refresh_from_db()
        self.assertNotEqual(
            self._test_document.document_type, test_document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self._test_document_types[1]
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_type_changed.id)

    def test_trashed_document_change_type_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_change_type
        )
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_change_type
        )

        test_document_type = self._test_document.document_type

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_change_type_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document.refresh_from_db()
        self.assertEqual(
            self._test_document.document_type, test_document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from ..permissions import permission_document_file_view

from .base import GenericDocumentViewTestCase
from .mixins.document_file_mixins import DocumentFilePageViewTestMixin


class DocumentFilePageViewTestCase(
    DocumentFilePageViewTestMixin, GenericDocumentViewTestCase
):
    def test_document_file_page_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_list_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_list_view()
        self.assertContains(
            response=response, status_code=200,
            text=str(self._test_document_file)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_list_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_rotate_left_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_rotate_left_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_rotate_left_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_rotate_left_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_rotate_left_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_rotate_left_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_rotate_right_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_rotate_right_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_rotate_right_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_rotate_right_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_rotate_right_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_rotate_right_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_view(
            document_file_page=self._test_document_file.pages.first()
        )
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_view(
            document_file_page=self._test_document_file.pages.first()
        )
        self.assertContains(
            response=response, status_code=200, text=str(
                self._test_document_file.pages.first()
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_view(
            document_file_page=self._test_document_file.pages.first()
        )
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_zoom_in_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_zoom_in_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_zoom_in_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_zoom_in_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_zoom_in_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_zoom_in_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_zoom_out_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_file_page_zoom_out_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_page_zoom_out_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._clear_events()

        response = self._request_test_document_file_page_zoom_out_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_file_page_zoom_out_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_file_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_page_zoom_out_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django import forms as django_forms

from mayan.apps.views.forms import Form

from ..classes import Layer
from ..forms import LayerTransformationForm


class DynamicTransformationFormClassMixin:
    def get_form_class(self):
        transformation_class = self.get_transformation_class()

        TransformationForm = transformation_class.get_form_class()

        if not TransformationForm and not getattr(transformation_class, 'template_name', None):
            # Transformation does not specify a form and does not have
            # a template either. Create a dynamic form based on the argument
            # list.
            class TransformationForm(Form):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    for argument in transformation_class.get_arguments():
                        self.fields[argument] = django_forms.CharField(required=False)

        class MergedTransformationForm(
            TransformationForm, LayerTransformationForm
        ):
            """Model form merged with the specific transformation fields."""
            view = self

        return MergedTransformationForm


class LayerViewMixin:
    def dispatch(self, request, *args, **kwargs):
        self.layer = self.get_layer()
        return super().dispatch(request=request, *args, **kwargs)

    def get_layer(self):
        return Layer.get(
            name=self.kwargs['layer_name']
        )



from ...literals import QUERY_PARAMETER_ANY_FIELD
from ...search_query_types import (
    QueryTypeExact, QueryTypeFuzzy, QueryTypeGreaterThan,
    QueryTypeGreaterThanOrEqual, QueryTypeLessThan, QueryTypeLessThanOrEqual,
    QueryTypePartial, QueryTypeRange, QueryTypeRangeExclusive,
    QueryTypeRegularExpression
)

from ..literals import TEST_OBJECT_INTEGER_VALUE

from .backend_mixins import BackendSearchTestMixin
from .base import SearchTestMixin, TestSearchObjectSimpleTestMixin


class BackendFieldTypeQueryTypeAnyTestCaseMixin:
    def test_search_field_type_any_search_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_any_search_empty_quoted(self):
        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_any_search_exact(self):
        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypeExact,
            value=self._test_object.char
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_any_search_exact_case_insensitive(self):
        self._test_object.char = self._test_object.char.upper()
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypeExact,
            value=self._test_object.char.lower()
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        self._test_object.char = self._test_object.char.lower()
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypeExact,
            value=self._test_object.char.upper()
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_any_search_partial(self):
        parts = self._test_object.char.split(' ')

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value=parts[0][:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value=parts[0][1:]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value=parts[0].upper()[:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value=parts[0].upper()[1:]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[0], parts[1]
                )
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[1], parts[0]
                )
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[0], parts[2]
                )
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value=self._test_object.email[1:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value=str(self._test_object.uuid).split('-')[0]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_any_search_exact_invalid(self):
        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypeExact,
            value=1.1
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_any_search_partial_invalid(self):
        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value='1970-01'
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypeAutoFieldTestCaseMixin:
    def test_search_field_type_autofield_search_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_autofield_search_exact(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeExact,
            value=self._test_object.id
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeExact,
            value=str(self._test_object.id)
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_autofield_search_greater_than(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeGreaterThan,
            value=self._test_object.id - 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_autofield_search_greater_than_or_equal(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeGreaterThanOrEqual,
            value=self._test_object.id - 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeGreaterThanOrEqual,
            value=self._test_object.id
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_autofield_search_less_than(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeLessThan,
            value=self._test_object.id + 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_autofield_search_less_than_or_equal(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeLessThanOrEqual,
            value=self._test_object.id + 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeLessThanOrEqual,
            value=self._test_object.id
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_autofield_search_range(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.id - 1, self._test_object.id + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.id, self._test_object.id + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.id - 1, self._test_object.id
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.id, self._test_object.id
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_autofield_search_range_exclusive(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.id - 1, self._test_object.id + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.id, self._test_object.id + 1
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.id - 1, self._test_object.id
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.id, self._test_object.id
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_autofield_search_range_exclusive_invalid(self):
        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRangeExclusive,
            value='INVALID'
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='id',
            query_type=QueryTypeRangeExclusive,
            value=True
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypeBooleanTestCaseMixin:
    def test_search_field_type_boolean_search_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.boolean = False
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_boolean_search_exact(self):
        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value=True
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value='true'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value='TRUE'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value=False
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value='false'
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value='FALSE'
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.boolean = False
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='boolean',
            query_type=QueryTypeExact,
            value='FALSE'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)


class BackendFieldTypeQueryTypeCharTestCaseMixin:
    def test_search_field_type_char_search_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.char = ''
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_char_search_empty_quoted(self):
        id_list = self._do_backend_search(
            field_name='char',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.char = ''
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='char',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_char_search_exact_accent(self):
        self._test_object.char = 'café'
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeExact,
            value='cafe'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_char_search_exact(self):
        parts = self._test_object.char.split(' ')

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeExact,
            value=parts[0]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeExact,
            value=''.join(
                (
                    parts[0], parts[1]
                )
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeExact,
            value=parts[0][1:]
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='char',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=self._test_object.char
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=' '.join(
                (
                    parts[2], parts[1], parts[0]
                )
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_char_search_fuzzy(self):
        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeFuzzy,
            value='chra'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            is_quoted_value=True,
            query_type=QueryTypeFuzzy,
            value='test chra'
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_char_search_partial(self):
        parts = self._test_object.char.split(' ')

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypePartial,
            value=parts[0][:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypePartial,
            value=parts[0][1:]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypePartial,
            value=parts[0].upper()[:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypePartial,
            value=parts[0].upper()[1:]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[0], parts[1]
                )
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[1], parts[0]
                )
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='char',
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[0], parts[2]
                )
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_char_search_regular_expression(self):
        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeRegularExpression,
            value='c.*r'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeRegularExpression,
            value='(test|INVALID)'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_char_search_invalid(self):
        id_list = self._do_backend_search(
            field_name='char',
            query_type=QueryTypeExact,
            value=99
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypeDateTimeTestCaseMixin:
    def test_search_field_type_datetime_search_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.datetime = None
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_datetime_search_empty_quoted(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.datetime = None
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='datetime',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_datetime_search_exact(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeExact,
            value='{year}-{month}-{day}T{hour}:{minute}:{second}'.format(
                year=self._test_object.datetime.year,
                month=self._test_object.datetime.month,
                day=self._test_object.datetime.day,
                hour=self._test_object.datetime.hour,
                minute=self._test_object.datetime.minute,
                second=self._test_object.datetime.second
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_greater_than(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeGreaterThan, value='{year}-{month}'.format(
                year=self._test_object.datetime.year - 1,
                month=self._test_object.datetime.month
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_greater_than_humanized_quoted(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            is_quoted_value=True,
            query_type=QueryTypeGreaterThan,
            value='last month'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_greater_than_humanized_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeGreaterThan,
            value='last month'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_greater_than_or_equal(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeGreaterThanOrEqual,
            value='{year}-{month}'.format(
                year=self._test_object.datetime.year - 1,
                month=self._test_object.datetime.month
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_less_than(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeLessThan, value='{year}-{month}'.format(
                year=self._test_object.datetime.year + 1,
                month=self._test_object.datetime.month
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_less_than_humanized(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeLessThan, value='tomorrow'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_less_than_or_equal(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeLessThanOrEqual,
            value='{year}-{month}'.format(
                year=self._test_object.datetime.year + 1,
                month=self._test_object.datetime.month
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_range(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeRange,
            value='{year_low}-{month_low}..{year_high}-{month_high}'.format(
                year_low=self._test_object.datetime.year - 1,
                month_low=self._test_object.datetime.month,
                year_high=self._test_object.datetime.year + 1,
                month_high=self._test_object.datetime.month
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_range_humanized(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            is_quoted_value=True,
            query_type=QueryTypeRange,
            value='last year..tomorrow'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='datetime',
            is_quoted_value=True,
            query_type=QueryTypeRange,
            value='yesterday..in two months'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_range_humanized_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeRange,
            value='last year..tomorrow'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_range_exclusive(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeRangeExclusive,
            value='{year_low}-{month_low}..{year_high}-{month_high}'.format(
                year_low=self._test_object.datetime.year - 1,
                month_low=self._test_object.datetime.month,
                year_high=self._test_object.datetime.year + 1,
                month_high=self._test_object.datetime.month
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_datetime_search_range_invalid(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeRangeExclusive,
            value=11111
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeRangeExclusive,
            value='invalid'
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='datetime',
            query_type=QueryTypeRangeExclusive,
            value=True
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypeEmailTestCaseMixin:
    def test_search_field_type_email_search_exact_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.email = ''
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_email_search_exact_empty_quoted(self):
        id_list = self._do_backend_search(
            field_name='email',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.email = ''
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='email',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_email_search_exact(self):
        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypeExact,
            value=self._test_object.email
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_email_search_partial(self):
        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypePartial,
            value='user'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypePartial,
            value='user@'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypePartial,
            value='example.org'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypePartial,
            value=self._test_object.email[1:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_email_search_exact_invalid(self):
        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypeExact,
            value=99
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='email',
            query_type=QueryTypeExact,
            value=True
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypeIntegerTestCaseMixin:
    def test_search_field_type_integer_null_search_exact_empty_non_quoted(self):
        self._test_object_integer_set = False

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.integer = TEST_OBJECT_INTEGER_VALUE
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_integer_null_search_exact_empty_quoted(self):
        self._test_object_integer_set = False

        id_list = self._do_backend_search(
            field_name='integer',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.integer = TEST_OBJECT_INTEGER_VALUE
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='integer',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_integer_search_exact(self):
        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeExact,
            value=self._test_object.integer
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeExact,
            value=str(self._test_object.integer)
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        # Test valid value ranges.
        self._test_object.integer = -2 ** 31
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeExact,
            value=str(self._test_object.integer)
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        self._test_object.integer = 2 ** 31 - 1
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeExact,
            value=str(self._test_object.integer)
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_integer_search_greater_than(self):
        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeGreaterThan,
            value=self._test_object.integer - 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_integer_search_greater_than_or_equal(self):
        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeGreaterThanOrEqual,
            value=self._test_object.integer - 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeGreaterThanOrEqual,
            value=self._test_object.integer
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_integer_search_less_than(self):
        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeLessThan,
            value=self._test_object.integer + 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_integer_search_less_than_or_equal(self):
        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeLessThanOrEqual,
            value=self._test_object.integer + 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeLessThanOrEqual,
            value=self._test_object.integer
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_integer_search_range(self):
        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.integer - 1, self._test_object.integer + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.integer, self._test_object.integer + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.integer - 1, self._test_object.integer
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.integer, self._test_object.integer
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_integer_search_range_exclusive(self):
        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.integer - 1, self._test_object.integer + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.integer, self._test_object.integer + 1
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.integer - 1, self._test_object.integer
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.integer, self._test_object.integer
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_integer_search_range_exclusive_invalid(self):
        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRangeExclusive,
            value='INVALID'
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='integer',
            query_type=QueryTypeRangeExclusive,
            value=True
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypePositiveIntegerTestCaseMixin:
    def test_search_field_type_positiveinteger_search_exact_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_positiveinteger_search_exact(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeExact,
            value=self._test_object.positiveinteger
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeExact,
            value=str(self._test_object.positiveinteger)
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        # Test valid value range.
        self._test_object.positiveinteger = 2 ** 31 - 1
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeExact,
            value=self._test_object.positiveinteger
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_positiveinteger_search_greater_than(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeGreaterThan,
            value=self._test_object.positiveinteger - 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_positiveinteger_search_greater_than_or_equal(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeGreaterThanOrEqual,
            value=self._test_object.positiveinteger - 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeGreaterThanOrEqual,
            value=self._test_object.positiveinteger
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_positiveinteger_search_less_than(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeLessThan,
            value=self._test_object.positiveinteger + 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_positiveinteger_search_less_than_or_equal(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeLessThanOrEqual,
            value=self._test_object.positiveinteger + 1
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeLessThanOrEqual,
            value=self._test_object.positiveinteger
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_positiveinteger_search_range(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.positiveinteger - 1, self._test_object.positiveinteger + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.positiveinteger, self._test_object.positiveinteger + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.positiveinteger - 1, self._test_object.positiveinteger
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRange,
            value='{}..{}'.format(
                self._test_object.positiveinteger, self._test_object.positiveinteger
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_positiveinteger_search_range_exclusive(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.positiveinteger - 1, self._test_object.positiveinteger + 1
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.positiveinteger, self._test_object.positiveinteger + 1
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.positiveinteger - 1, self._test_object.positiveinteger
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRangeExclusive,
            value='{}..{}'.format(
                self._test_object.positiveinteger, self._test_object.positiveinteger
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_positiveinteger_search_range_exclusive_invalid(self):
        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRangeExclusive,
            value='INVALID'
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='positiveinteger',
            query_type=QueryTypeRangeExclusive,
            value=True
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypeTextTestCaseMixin:
    def test_search_field_type_text_search_exact_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_text_search_exact_empty_quoted(self):
        id_list = self._do_backend_search(
            field_name='text',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

        self._test_object.text = ''
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='text',
            is_quoted_value=True,
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_text_search_exact(self):
        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypeExact,
            value=self._test_object.text
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_text_search_fuzzy(self):
        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypeFuzzy,
            value='tetx'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='text',
            is_quoted_value=True,
            query_type=QueryTypeFuzzy,
            value='test tetx'
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_text_search_partial(self):
        parts = self._test_object.text.split(' ')

        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypePartial,
            value=parts[0][:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypePartial,
            value=parts[0][1:]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypePartial,
            value=parts[0].upper()[:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypePartial,
            value=parts[0].upper()[1:]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='text',
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[0], parts[1]
                )
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='text',
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[1], parts[0]
                )
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='text',
            is_quoted_value=True,
            query_type=QueryTypePartial,
            value=' '.join(
                (
                    parts[0], parts[2]
                )
            )
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_text_search_regular_expression(self):
        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypeRegularExpression,
            value='t.*t'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypeRegularExpression,
            value='(text|INVALID)'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_text_search_regular_expression_invalid(self):
        id_list = self._do_backend_search(
            field_name='text',
            query_type=QueryTypeRegularExpression,
            value=99
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypeUUIDTestCaseMixin:
    def test_search_field_type_uuid_search_exact_empty_non_quoted(self):
        id_list = self._do_backend_search(
            field_name='uuid',
            query_type=QueryTypeExact,
            value=''
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_uuid_search_exact(self):
        id_list = self._do_backend_search(
            field_name='uuid',
            query_type=QueryTypeExact,
            value=self._test_object.uuid
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_uuid_search_partial(self):
        id_list = self._do_backend_search(
            field_name='uuid',
            query_type=QueryTypePartial,
            value=str(self._test_object.uuid)[1:-1]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='uuid',
            query_type=QueryTypePartial,
            value=str(self._test_object.uuid).split('-')[0]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_uuid_search_regular_expression(self):
        parts = str(self._test_object.uuid).split('-')

        id_list = self._do_backend_search(
            field_name='uuid',
            query_type=QueryTypeRegularExpression,
            value='{}.*{}.*'.format(
                parts[0][:5], parts[0][-1]
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='uuid',
            query_type=QueryTypeRegularExpression,
            value='({}|INVALID)'.format(
                '{}.*{}.*'.format(
                    parts[0][:5], parts[0][-1]
                )
            )
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_search_field_type_uuid_search_exact_invalid(self):
        id_list = self._do_backend_search(
            field_name='uuid',
            query_type=QueryTypeExact,
            value=True
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_search_field_type_uuid_search_regular_expression_invalid(self):
        id_list = self._do_backend_search(
            field_name='uuid',
            query_type=QueryTypeRegularExpression,
            value=True
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendFieldTypeQueryTypeTestCaseMixin(
    BackendFieldTypeQueryTypeAnyTestCaseMixin,
    BackendFieldTypeQueryTypeAutoFieldTestCaseMixin,
    BackendFieldTypeQueryTypeBooleanTestCaseMixin,
    BackendFieldTypeQueryTypeCharTestCaseMixin,
    BackendFieldTypeQueryTypeDateTimeTestCaseMixin,
    BackendFieldTypeQueryTypeEmailTestCaseMixin,
    BackendFieldTypeQueryTypeIntegerTestCaseMixin,
    BackendFieldTypeQueryTypePositiveIntegerTestCaseMixin,
    BackendFieldTypeQueryTypeTextTestCaseMixin,
    BackendFieldTypeQueryTypeUUIDTestCaseMixin,
    BackendSearchTestMixin, TestSearchObjectSimpleTestMixin,
    SearchTestMixin
):
    """
    Consolidated backend test field type and query type case mixin.
    """



from mayan.apps.document_indexing.models.index_instance_models import IndexInstanceNode
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.mixins.document_version_mixins import DocumentVersionTestMixin

from .literals import (
    TEST_DOCUMENT_VERSION_OCR_INDEX_NODE_TEMPLATE,
    TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
)
from .mixins import DocumentVersionOCRTestMixin


class DocumentVersionOCRIndexingTestCase(
    DocumentVersionOCRTestMixin, DocumentVersionTestMixin,
    IndexTemplateTestMixin, GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_DOCUMENT_VERSION_OCR_INDEX_NODE_TEMPLATE
    auto_create_test_document_stub = True
    auto_create_test_document_version = True
    auto_create_test_document_version_page = True
    auto_upload_test_document = False

    def test_indexing_not_document_version_ocr(self):
        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document
            ).exists()
        )

    def test_indexing_document_version_ocr_delete(self):
        self._create_test_document_version_ocr_content()
        value = ' '.join(
            self._test_document.version_active.ocr_content()
        )

        self._do_test_document_version_ocr_content_delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )

    def test_indexing_document_version_ocr_edit(self):
        self._create_test_document_version_ocr_content()
        value = ' '.join(
            self._test_document.version_active.ocr_content()
        )

        self._test_document_version_page.ocr_content.content = TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
        self._test_document_version_page.ocr_content.save()

        value_edited = ' '.join(
            self._test_document.version_active.ocr_content()
        )

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )
        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value_edited
            ).exists()
        )

    def test_indexing_document_version_ocr_finished(self):
        self._create_test_document_version_ocr_content()
        value = ' '.join(
            self._test_document.version_active.ocr_content()
        )

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )



from django.contrib import messages
from django.utils.translation import ugettext_lazy as _, ungettext

from mayan.apps.views.generics import (
    ConfirmView, MultipleObjectConfirmActionView, SingleObjectDetailView,
    SingleObjectListView
)
from mayan.apps.views.view_mixins import (
    ContentTypeViewMixin, ExternalObjectViewMixin
)

from .forms import CacheDetailForm, CachePartitionDetailForm
from .icons import (
    icon_cache_detail, icon_cache_list, icon_cache_partition_detail,
    icon_cache_partition_purge, icon_cache_purge
)
from .models import Cache, CachePartition
from .permissions import (
    permission_cache_partition_purge, permission_cache_purge,
    permission_cache_view
)

from .tasks import task_cache_partition_purge, task_cache_purge


class CacheDetailView(SingleObjectDetailView):
    form_class = CacheDetailForm
    form_extra_kwargs = {
        'extra_fields': [
            {
                'field': 'label'
            },
            {
                'field': 'get_maximum_size_display'
            },
            {
                'field': 'get_total_size_display'
            }
        ]
    }
    model = Cache
    object_permission = permission_cache_view
    pk_url_kwarg = 'cache_id'
    view_icon = icon_cache_detail

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Details of cache: %s') % self.object
        }


class CacheListView(SingleObjectListView):
    model = Cache
    object_permission = permission_cache_view
    view_icon = icon_cache_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'title': _('File caches list')
        }


class CachePartitionDetailView(SingleObjectDetailView):
    form_class = CachePartitionDetailForm
    form_extra_kwargs = {
        'extra_fields': [
            {
                'field': 'get_total_size_display'
            }
        ]
    }
    model = CachePartition
    object_permission = permission_cache_view
    pk_url_kwarg = 'cache_partition_id'
    view_icon = icon_cache_partition_detail

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Details of cache partition: %s') % self.object
        }


class CachePartitionPurgeView(
    ContentTypeViewMixin, ExternalObjectViewMixin, ConfirmView
):
    external_object_permission = permission_cache_partition_purge
    external_object_pk_url_kwarg = 'object_id'
    view_icon = icon_cache_partition_purge

    def get_external_object_queryset(self):
        return self.get_content_type().get_all_objects_for_this_type()

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                'Purge cache partitions of "%s"?'
            ) % self.external_object
        }

    def view_action(self, form=None):
        for cache_partition in self.external_object.get_cache_partitions():
            task_cache_partition_purge.apply_async(
                kwargs={
                    'content_type_id': self.get_content_type().pk,
                    'cache_partition_id': cache_partition.pk,
                    'object_id': self.external_object.pk,
                    'user_id': self.request.user.pk
                }
            )

        messages.success(
            message=_(
                'Object cache partitions submitted for purging.'
            ), request=self.request
        )


class CachePurgeView(MultipleObjectConfirmActionView):
    model = Cache
    object_permission = permission_cache_purge
    pk_url_kwarg = 'cache_id'
    success_message_plural = _('%(count)d caches submitted for purging.')
    success_message_singular = _('%(count)d cache submitted for purging.')
    view_icon = icon_cache_purge

    def get_extra_context(self):
        queryset = self.object_list

        result = {
            'title': ungettext(
                singular='Submit the selected cache for purging?',
                plural='Submit the selected caches for purging?',
                number=queryset.count()
            )
        }

        if queryset.count() == 1:
            result['object'] = queryset.first()

        return result

    def object_action(self, form, instance):
        task_cache_purge.apply_async(
            kwargs={'cache_id': instance.pk, 'user_id': self.request.user.pk}
        )



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..links import link_transformation_delete

from .mixins import TransformationViewTestMixin


class TransformationLinkConditionTestCase(
    TransformationViewTestMixin, GenericDocumentViewTestCase
):
    auto_login_super_user = True
    create_test_case_super_user = True
    create_test_case_user = False

    def test_transformation_condition_link_view_with_super_user(self):
        self._create_test_transformation()

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertContains(
            response=response, text=str(self._test_transformation_object),
            status_code=200
        )
        self.assertContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class TransformationLinkDisplayTestCase(
    TransformationViewTestMixin, GenericDocumentViewTestCase
):
    def test_transformation_delete_link_view_with_view_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=self._test_layer_permission_view
        )

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertContains(
            response=response, text=str(self._test_transformation_object),
            status_code=200
        )
        self.assertContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=200
        )
        self.assertNotContains(
            response=response,
            text=link_transformation_delete.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_transformation_delete_link_view_with_all_access(self):
        self._create_test_transformation()

        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=self._test_layer_permission_delete
        )
        self.grant_access(
            obj=self._test_transformation_object_parent,
            permission=self._test_layer_permission_view
        )

        self._clear_events()

        response = self._request_transformation_list_view()
        self.assertContains(
            response=response, text=str(self._test_transformation_object),
            status_code=200
        )
        self.assertContains(
            response=response,
            text=self._test_transformation.get_transformation_class().label,
            status_code=200
        )
        self.assertContains(
            response=response,
            text=link_transformation_delete.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from unittest import mock

from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH

from .literals import TEST_SOURCE_BACKEND_PATH_INTERACTIVE_ACTION
from .mixins.interactive_mixins import InteractiveSourceBackendTestMixin


class SourceBackendCallbackTestCase(
    InteractiveSourceBackendTestMixin, GenericDocumentTestCase
):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_INTERACTIVE_ACTION
    auto_upload_test_document = False

    @mock.patch(
        target='mayan.apps.sources.models.Source.callback_post_document_create'
    )
    def test_callback_post_document_create_on_document_upload(
        self, mocked_source_method
    ):
        document_count = Document.objects.count()

        self._clear_events()

        with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={'file_object': file_object}
            )

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(mocked_source_method.call_count, 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

    @mock.patch(
        target='mayan.apps.sources.models.Source.callback_post_document_file_upload'
    )
    def test_callback_post_document_file_upload_on_document_upload(
        self, mocked_source_method
    ):
        document_count = Document.objects.count()

        self._clear_events()

        with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={'file_object': file_object}
            )

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(mocked_source_method.call_count, 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

    @mock.patch(
        target='mayan.apps.sources.models.Source.callback_post_document_file_upload'
    )
    def test_callback_post_document_file_upload_on_document_file_upload(
        self, mocked_source_method
    ):
        self._create_test_document_stub()

        document_count = Document.objects.count()
        document_file_count = DocumentFile.objects.count()

        self._clear_events()

        with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_file_upload',
                extra_data={'file_object': file_object}
            )

        self.assertEqual(
            Document.objects.count(), document_count
        )
        self.assertEqual(
            DocumentFile.objects.count(), document_file_count + 1
        )

        self.assertEqual(mocked_source_method.call_count, 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, test_document)
        self.assertEqual(events[0].actor, test_document_file)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, test_document_version_page)
        self.assertEqual(events[3].target, test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )



from ...models.workflow_state_models import (
    WorkflowState, WorkflowStateRuntimeProxy
)

from ..literals import (
    TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION,
    TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
    TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED
)

from .workflow_template_mixins import WorkflowTemplateTestMixin


class WorkflowTemplateStateTestMixin(WorkflowTemplateTestMixin):
    _test_object_model = WorkflowState
    _test_object_name = '_test_workflow_template_state'
    auto_create_test_workflow_template_state = False

    def setUp(self):
        super().setUp()
        self._test_workflow_template_states = []

        if self.auto_create_test_workflow_template_state:
            self._create_test_workflow_template_state()

    def _create_test_workflow_template_state(self):
        total_test_workflow_template_states = len(
            self._test_workflow_template_states
        )
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
            total_test_workflow_template_states
        )
        initial = not self._test_workflow_template.states.exists()

        self._test_workflow_template_state = self._test_workflow_template.states.create(
            completion=TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION,
            initial=initial, label=label
        )
        self._test_workflow_template_states.append(
            self._test_workflow_template_state
        )
        self._test_workflow_template_state_runtime_proxy = WorkflowStateRuntimeProxy.objects.get(
            pk=self._test_workflow_template_state.pk
        )
        self._test_workflow_template_state_runtime_proxies.append(
            self._test_workflow_template_state_runtime_proxy
        )


class WorkflowTemplateStateAPIViewTestMixin(WorkflowTemplateStateTestMixin):
    def _request_test_workflow_template_state_create_api_view(self):
        self._test_object_track()

        response = self.post(
            viewname='rest_api:workflow-template-state-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }, data={
                'completion': TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION,
                'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL
            }
        )

        self._test_object_set()

        return response

    def _request_test_workflow_template_state_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-state-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_state.pk
            }
        )

    def _request_test_workflow_template_state_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-state-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_state.pk
            }
        )

    def _request_test_workflow_template_state_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-state-list', kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_state_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-state-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_state.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED
            }
        )

    def _request_test_workflow_template_state_edit_put_api_view(self):
        return self.put(
            viewname='rest_api:workflow-template-state-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_state_id': self._test_workflow_template_state.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED
            }
        )


class WorkflowTemplateStateViewTestMixin(WorkflowTemplateStateTestMixin):
    def _request_test_workflow_template_state_create_view(self, extra_data=None):
        self._test_object_track()

        data = {
            'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
            'completion': TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION,
        }
        if extra_data:
            data.update(extra_data)

        response = self.post(
            viewname='document_states:workflow_template_state_create',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }, data=data
        )

        self._test_object_set()

        return response

    def _request_test_workflow_template_state_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_state_delete',
            kwargs={
                'workflow_template_state_id': self._test_workflow_template_states[0].pk
            }
        )

    def _request_test_workflow_template_state_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_state_edit',
            kwargs={
                'workflow_template_state_id': self._test_workflow_template_states[0].pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED
            }
        )

    def _request_test_workflow_template_state_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_state_list',
            kwargs={'workflow_template_id': self._test_workflow_template.pk}
        )



import logging
import operator

from django.conf import settings

from rest_framework.filters import BaseFilterBackend

from .exceptions import DynamicSearchException
from .literals import QUERY_PARAMETER_ANY_FIELD
from .search_backends import SearchBackend
from .search_models import SearchModel

logger = logging.getLogger(name=__name__)


class RESTAPISearchFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        if not getattr(view, 'search_disable_list_filtering', False):
            search_model = self.get_search_model(queryset=queryset)
            if search_model:
                query_dict = request.GET.dict().copy()
                query_dict.update(
                    request.POST.dict()
                )

                search_model_fields = list(
                    map(
                        operator.itemgetter(0),
                        search_model.get_search_field_choices()
                    )
                )

                search_model_fields.append(QUERY_PARAMETER_ANY_FIELD)

                valid_search_models_query_dict_keys = set(
                    query_dict.keys()
                ).intersection(
                    set(search_model_fields)
                )

                query_dict_cleaned = {
                    key: query_dict[key] for key in valid_search_models_query_dict_keys
                }

                if query_dict_cleaned:
                    try:
                        queryset_search = SearchBackend.get_instance().search(
                            search_model=search_model,
                            query=query_dict_cleaned, user=request.user
                        )
                    except DynamicSearchException as exception:
                        if settings.DEBUG or settings.TESTING:
                            raise

                        logger.error(
                            'Error performing REST API list search '
                            'filtering; %s', exception
                        )
                        return search_model.model._meta.default_manager.none()
                    else:
                        return queryset.filter(pk__in=queryset_search)
                else:
                    return queryset
            else:
                return queryset
        else:
            return queryset

    def get_search_model(self, queryset):
        try:
            model = queryset.model
        except AttributeError:
            return
        else:
            try:
                return SearchModel.get_for_model(instance=model)
            except KeyError:
                return



from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ...classes import BaseDocumentFilenameGenerator
from ...models.document_type_models import DocumentType, DocumentTypeFilename

from ..literals import (
    TEST_DOCUMENT_TYPE_DELETE_PERIOD, TEST_DOCUMENT_TYPE_DELETE_TIME_UNIT,
    TEST_DOCUMENT_TYPE_LABEL, TEST_DOCUMENT_TYPE_LABEL_EDITED,
    TEST_DOCUMENT_TYPE_QUICK_LABEL, TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED
)


class DocumentTypeTestMixin(TestMixinObjectCreationTrack):
    _test_object_model = DocumentType
    _test_object_name = '_test_document_type'
    auto_create_test_document_type = True
    auto_delete_test_document_type = True

    def setUp(self):
        super().setUp()

        self._test_document_types = []

        if self.auto_create_test_document_type:
            self._create_test_document_type()

    def tearDown(self):
        if self.auto_delete_test_document_type:
            for document_type in DocumentType.objects.all():
                document_type.delete()
        super().tearDown()

    def _create_test_document_type(self, label=None):
        label = label or '{}_{}'.format(
            TEST_DOCUMENT_TYPE_LABEL, len(self._test_document_types)
        )

        self._test_document_type = DocumentType.objects.create(label=label)
        self._test_document_types.append(self._test_document_type)


class DocumentTypeQuickLabelTestMixin(DocumentTypeTestMixin):
    _test_object_model = DocumentTypeFilename
    _test_object_name = '_test_document_type_quick_label'

    def _create_test_document_type_quick_label(self):
        self._test_document_type_quick_label = self._test_document_type.filenames.create(
            filename=TEST_DOCUMENT_TYPE_QUICK_LABEL
        )


class DocumentQuickLabelViewTestMixin(DocumentTypeQuickLabelTestMixin):
    def _request_test_document_quick_label_edit_view(self, extra_data=None):
        data = {
            'document_type_available_filenames': self._test_document_type_quick_label.pk,
            'label': ''
            # View needs at least an empty label for quick
            # label to work. Cause is unknown.
        }
        data.update(
            extra_data or {}
        )

        return self.post(
            viewname='documents:document_properties_edit', kwargs={
                'document_id': self._test_document.pk
            }, data=data
        )


class DocumentTypeAPIViewTestMixin(DocumentTypeTestMixin):
    def _request_test_document_type_create_api_view(self):
        self._test_object_track()

        response = self.post(
            viewname='rest_api:documenttype-list', data={
                'label': TEST_DOCUMENT_TYPE_LABEL
            }
        )

        self._test_object_set()

        return response

    def _request_test_document_type_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documenttype-detail', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )

    def _request_test_document_type_detail_api_view(self):
        return self.get(
            viewname='rest_api:documenttype-detail', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )

    def _request_test_document_type_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:documenttype-detail', kwargs={
                'document_type_id': self._test_document_type.pk
            }, data={'label': TEST_DOCUMENT_TYPE_LABEL_EDITED}
        )

    def _request_test_document_type_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:documenttype-detail', kwargs={
                'document_type_id': self._test_document_type.pk
            }, data={'label': TEST_DOCUMENT_TYPE_LABEL_EDITED}
        )

    def _request_test_document_type_list_api_view(self):
        return self.get(viewname='rest_api:documenttype-list')


class DocumentTypeFilenameGeneratorViewTestMixin(DocumentTypeTestMixin):
    def _request_test_document_type_filename_generator_get_view(self):
        return self.get(
            viewname='documents:document_type_filename_generator', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )

    def _request_test_document_type_filename_generator_post_view(self):
        return self.post(
            viewname='documents:document_type_filename_generator', kwargs={
                'document_type_id': self._test_document_type.pk
            }, data={
                'filename_generator_backend': BaseDocumentFilenameGenerator.get_default()
            }
        )


class DocumentTypeQuickLabelAPIViewTestMixin(DocumentTypeQuickLabelTestMixin):
    def _request_test_document_type_quick_label_create_api_view(self):
        self._test_object_track()

        response = self.post(
            viewname='rest_api:documenttype-quicklabel-list', kwargs={
                'document_type_id': self._test_document_type.pk
            }, data={
                'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL
            }
        )

        self._test_object_set()

        return response

    def _request_test_document_type_quick_label_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documenttype-quicklabel-detail', kwargs={
                'document_type_id': self._test_document_type.pk,
                'document_type_quick_label_id': self._test_document_type_quick_label.pk
            }
        )

    def _request_test_document_type_quick_label_detail_api_view(self):
        return self.get(
            viewname='rest_api:documenttype-quicklabel-detail', kwargs={
                'document_type_id': self._test_document_type.pk,
                'document_type_quick_label_id': self._test_document_type_quick_label.pk
            }
        )

    def _request_test_document_type_quick_label_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:documenttype-quicklabel-detail', kwargs={
                'document_type_id': self._test_document_type.pk,
                'document_type_quick_label_id': self._test_document_type_quick_label.pk
            }, data={'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED}
        )

    def _request_test_document_type_quick_label_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:documenttype-quicklabel-detail', kwargs={
                'document_type_id': self._test_document_type.pk,
                'document_type_quick_label_id': self._test_document_type_quick_label.pk
            }, data={'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED}
        )

    def _request_test_document_type_quick_label_list_api_view(self):
        return self.get(
            viewname='rest_api:documenttype-quicklabel-list', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )


class DocumentTypeQuickLabelViewTestMixin(DocumentTypeQuickLabelTestMixin):
    def _request_test_quick_label_create_view(self):
        return self.post(
            viewname='documents:document_type_filename_create', kwargs={
                'document_type_id': self._test_document_type.pk
            }, data={
                'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL
            }
        )

    def _request_test_quick_label_delete_view(self):
        return self.post(
            viewname='documents:document_type_filename_delete', kwargs={
                'document_type_filename_id': self._test_document_type_quick_label.pk
            }
        )

    def _request_test_quick_label_edit_view(self):
        return self.post(
            viewname='documents:document_type_filename_edit', kwargs={
                'document_type_filename_id': self._test_document_type_quick_label.pk
            }, data={
                'filename': TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED
            }
        )

    def _request_test_quick_label_list_view(self):
        return self.get(
            viewname='documents:document_type_filename_list', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )


class DocumentTypeRetentionPoliciesViewTestMixin(DocumentTypeTestMixin):
    def _request_test_document_type_retention_policies_get_view(self):
        return self.get(
            viewname='documents:document_type_retention_policies', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )

    def _request_test_document_type_retention_policies_post_view(
        self, extra_data=None
    ):
        data = {
            'document_stub_expiration_interval': self._test_document_type.document_stub_expiration_interval,
            'document_stub_pruning_enabled': self._test_document_type.document_stub_pruning_enabled
        }

        if extra_data is not None:
            data.update(extra_data)

        return self.post(
            viewname='documents:document_type_retention_policies',
            kwargs={'document_type_id': self._test_document_type.pk},
            data=data
        )


class DocumentTypeViewTestMixin(DocumentTypeTestMixin):
    def _request_test_document_type_create_view(self):
        return self.post(
            viewname='documents:document_type_create',
            data={
                'label': TEST_DOCUMENT_TYPE_LABEL,
                'delete_time_period': TEST_DOCUMENT_TYPE_DELETE_PERIOD,
                'delete_time_unit': TEST_DOCUMENT_TYPE_DELETE_TIME_UNIT
            }
        )

    def _request_test_document_type_delete_view(self):
        return self.post(
            viewname='documents:document_type_delete', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )

    def _request_test_document_type_edit_view(self):
        return self.post(
            viewname='documents:document_type_edit', kwargs={
                'document_type_id': self._test_document_type.pk
            }, data={
                'label': TEST_DOCUMENT_TYPE_LABEL_EDITED
            }
        )

    def _request_test_document_type_list_view(self):
        return self.get(viewname='documents:document_type_list')



from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH
from mayan.apps.sources.tests.literals import TEST_CASE_INTERFACE_NAME_MODEL
from mayan.apps.sources.tests.mixins.base_mixins import SourceTestMixin

from .literals import TEST_SOURCE_BACKEND_PATH_COMPRESSED


class CompressedSourceTestMixin(SourceTestMixin):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_COMPRESSED
    _test_source_file_path = TEST_FILE_SMALL_PATH

    def _get_test_source_backend_data(self, action_name, interface_name):
        result = super()._get_test_source_backend_data(
            action_name=action_name, interface_name=interface_name
        )

        if action_name == 'document_upload':
            if interface_name == TEST_CASE_INTERFACE_NAME_MODEL:
                result.update(
                    {
                        'document_type': self._test_document_type
                    }
                )

        return result



from django.utils.translation import ugettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField
from mayan.apps.storage.models import SharedUploadedFile

from ..models.document_models import Document
from ..models.document_type_models import DocumentType
from ..permissions import permission_document_change_type
from ..tasks import task_document_file_upload

from .document_file_serializers import DocumentFileSerializer
from .document_type_serializers import DocumentTypeSerializer
from .document_version_serializers import DocumentVersionSerializer


class DocumentFileActionSerializer(serializers.Serializer):
    id = serializers.CharField(
        label=_('ID'), read_only=True, source='backend_id'
    )
    label = serializers.CharField(
        label=_('Label'), read_only=True
    )


class DocumentSerializer(serializers.HyperlinkedModelSerializer):
    document_type = DocumentTypeSerializer(
        label=_('Document type'), read_only=True
    )
    document_type_id = serializers.IntegerField(
        help_text=_('Document type ID for the new document.'),
        label=_('Document type ID'), write_only=True
    )
    document_change_type_url = serializers.HyperlinkedIdentityField(
        label=_('Document change type URL'), lookup_url_kwarg='document_id',
        view_name='rest_api:document-change-type'
    )
    file_latest = DocumentFileSerializer(
        label=_('File latest'), many=False, read_only=True
    )
    file_list_url = serializers.HyperlinkedIdentityField(
        label=_('File list URL'), lookup_url_kwarg='document_id',
        view_name='rest_api:documentfile-list'
    )
    version_active = DocumentVersionSerializer(
        label=_('Version active'), many=False, read_only=True
    )
    version_list_url = serializers.HyperlinkedIdentityField(
        label=_('Version list URL'), lookup_url_kwarg='document_id',
        view_name='rest_api:documentversion-list'
    )

    class Meta:
        create_only_fields = ('document_type_id',)
        extra_kwargs = {
            'url': {
                'label': _('URL'),
                'lookup_url_kwarg': 'document_id',
                'view_name': 'rest_api:document-detail'
            }
        }
        fields = (
            'datetime_created', 'description', 'document_change_type_url',
            'document_type', 'document_type_id', 'file_latest',
            'file_list_url', 'id', 'label', 'language', 'url', 'uuid',
            'version_active', 'version_list_url'
        )
        model = Document
        read_only_fields = (
            'datetime_created', 'document_change_type_url', 'document_type',
            'file_latest', 'file_list_url', 'id', 'uuid', 'url',
            'version_list_url'
        )


class DocumentChangeTypeSerializer(serializers.Serializer):
    document_type_id = FilteredPrimaryKeyRelatedField(
        label=_('Document type ID'), help_text=_(
            'Primary key of the document type into which the document '
            'will be changed.'
        ), source_permission=permission_document_change_type,
        source_queryset_method='get_document_type_queryset', write_only=True
    )

    def get_document_type_queryset(self):
        return DocumentType.objects.exclude(
            pk=self.context['object'].document_type.pk
        )


class DocumentUploadSerializer(DocumentSerializer):
    file = serializers.FileField(
        label=_('File'), write_only=True
    )

    def create(self, validated_data):
        file = validated_data.pop('file')
        validated_data['label'] = validated_data.get(
            'label', str(file)
        )
        user = validated_data['_instance_extra_data']['_event_actor']
        instance = super().create(validated_data=validated_data)

        shared_uploaded_file = SharedUploadedFile.objects.create(file=file)

        task_document_file_upload.apply_async(
            kwargs={
                'document_id': instance.pk,
                'shared_uploaded_file_id': shared_uploaded_file.pk,
                'user_id': user.pk
            }
        )

        return instance

    class Meta(DocumentSerializer.Meta):
        create_only_fields = ('document_type_id', 'file')
        fields = (
            'datetime_created', 'description', 'document_change_type_url',
            'document_type', 'document_type_id', 'file', 'file_list_url',
            'id', 'label', 'language', 'file_latest', 'pk', 'url', 'uuid',
            'version_active', 'version_list_url'
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import DEFAULT_AUTHENTICATION_OIDC_USER_PROFILE_URL

namespace = SettingNamespace(
    label=_('Authentication OIDC'), name='authentication_oidc'
)

setting_oidc_user_profile_url = namespace.add_setting(
    default=DEFAULT_AUTHENTICATION_OIDC_USER_PROFILE_URL,
    global_name='AUTHENTICATION_OIDC_USER_PROFILE_URL',
    help_text=_('URL for the OIDC user profile page.')
)
